/// @file    plugins/s3/S3.cpp
/// @brief   s3 plugin.
/// @author  Martin Philipp Hellmich <mhellmic@cern.ch>

#include <fcntl.h>

#include <dmlite/cpp/catalog.h>
#include <dmlite/cpp/poolmanager.h>
#include <dmlite/cpp/utils/urls.h>

#include "S3.h"
#include "S3Factory.h"

using namespace dmlite;

/* S3 pool handling */
S3PoolDriver::S3PoolDriver(S3Factory *factory) throw (DmException):
            stack(0x00),
            factory_(factory)
{
  // Nothing
}

S3PoolDriver::~S3PoolDriver()
{
  // Nothing
}

std::string S3PoolDriver::getImplId() const throw ()
{
  return std::string("S3PoolDriver");
}

void S3PoolDriver::setSecurityContext(const SecurityContext*) throw (DmException)
{
  // TODO
}

PoolHandler* S3PoolDriver::createPoolHandler(const std::string& poolName) throw (DmException)
{
  Pool meta = this->stack->getPoolManager()->getPool(poolName);
  
  int mode;
  if (meta.getString("mode").empty())
    mode = O_RDONLY;
  else if (meta.getString("mode") == "r")
    mode = O_RDONLY;
  else if (meta.getString("mode") == "w")
    mode = O_WRONLY;
  else if (meta.getString("mode") == "rw")
    mode = O_RDWR;
  else
    throw DmException(DMLITE_SYSERR(DMLITE_MALFORMED),
                      "Wrong mode '%s' configured in the database for the s3 pool %s",
                      meta.getString("mode").c_str(), poolName.c_str());

  S3PoolDetails pool;
  pool.set_host(meta.getString("hostname", "s3.amazonaws.com"));
  pool.set_redirect_port(meta.getLong("redirect_port", 80));
  pool.set_backend_port(meta.getLong("backend_port", 443));
  pool.set_backend_protocol(meta.getString("backend_protocol", "https"));
  pool.set_s3id(meta.getString("s3accesskeyid"));
  pool.set_s3key(meta.getString("s3secretaccesskey"));
  pool.set_bucketsalt(meta.getString("bucketsalt", "-dpm-forever"));
  pool.set_signedlinktimeout(meta.getLong("signedlinktimeout", 60));
  pool.set_usetorrent(meta.getBool("usetorrent", false));

  if (meta.getString("buckettype").empty())
    pool.set_buckettype(S3PoolDetails::SUBDOMAIN);
  else if (meta.getString("buckettype") == "subdomain")
    pool.set_buckettype(S3PoolDetails::SUBDOMAIN);
  else if (meta.getString("buckettype") == "path")
    pool.set_buckettype(S3PoolDetails::PATH);

  pool.set_mode(mode);
  
  return new S3PoolHandler(this->factory_, poolName, this->stack, pool);
}

void S3PoolDriver::toBeCreated(const Pool& pool) throw (DmException)
{
  // Nothing
}

void S3PoolDriver::justCreated(const Pool& pool) throw (DmException)
{
  // Nothing
}

void S3PoolDriver::update(const Pool& pool) throw (DmException)
{
  // Nothing
}

void S3PoolDriver::toBeDeleted(const Pool& pool) throw (DmException)
{
  // Nothing
}

void S3PoolDriver::setStackInstance(StackInstance* si) throw (DmException)
{
  this->stack = si;
}

S3PoolHandler::S3PoolHandler(S3Factory *factory, const std::string& poolName, StackInstance* si, S3PoolDetails pool):
  factory_(factory),
  poolName(poolName),
  stack(si),
  pool_(pool)
{
  this->conn_ = factory_->getConnection();
  this->conn_->setStackInstance(si);
}

S3PoolHandler::~S3PoolHandler()
{
  this->factory_->releaseConnection(this->conn_);
}


std::string S3PoolHandler::getPoolType(void) throw (DmException)
{
  return "s3";
}

std::string S3PoolHandler::getPoolName(void) throw (DmException)
{
  return this->poolName;
}

uint64_t S3PoolHandler::getTotalSpace(void) throw (DmException)
{
  return (uint64_t) POOL_SPACE;
}

uint64_t S3PoolHandler::getFreeSpace(void) throw (DmException)
{
  return (uint64_t) POOL_SPACE;
}

bool S3PoolHandler::poolIsAvailable(bool write) throw (DmException)
{
  int mode = this->pool_.mode();
  return (mode == O_RDWR) || (mode == O_RDONLY && !write) || (mode == O_WRONLY && write);
}

bool S3PoolHandler::replicaIsAvailable(const Replica& replica) throw (DmException)
{
  bool isAvailable = false;

  // if PENDING, check the file on S3
  switch (replica.status) {
  case Replica::kAvailable:
    isAvailable = true;
    break;
  case Replica::kBeingPopulated:
    isAvailable = 
          this->conn_->checkReplicaAvailability(replica, this->pool_);
    break;
  case Replica::kToBeDeleted:
    isAvailable = false;
    break;
  }

  return isAvailable;
}

Location S3PoolHandler::whereToRead(const Replica& replica) throw (DmException)
{
  Location rloc;
  time_t expiration = time(NULL) + static_cast<time_t>(this->pool_.signedlinktimeout());
  rloc = this->conn_->getQueryString(
//              replica.server,
//              replica.getUnsigned("port"),
              replica.getString("bucket"), //chooseBucket(replica)
              replica.rfn,
              expiration,
              this->pool_);

  return rloc;
}

void S3PoolHandler::removeReplica(const Replica& replica) throw (DmException)
{
  this->conn_->removeReplica(replica, this->pool_);
}

Location S3PoolHandler::whereToWrite(const std::string& fn) throw (DmException)
{
  // create database entry with state PENDING
  struct stat st = this->stack->getCatalog()->extendedStat(fn).stat;
  
  Replica replica;
  replica.fileid = st.st_ino;
  replica.status = Replica::kBeingPopulated;
  replica.type   = Replica::kPermanent;
  replica.rfn = fn;
  replica.server = this->pool_.host();
//  replica["port"] = this->pool_.port();
  replica["pool"] = this->poolName;
  replica["bucket"] = chooseBucket(replica);

  this->stack->getINode()->addReplica(replica);

  this->conn_->addPendingReplica(replica, this->pool_);

  // create PUT link
  Location rloc;
  time_t expiration = time(NULL) + static_cast<time_t>(this->pool_.signedlinktimeout());
  rloc = this->conn_->putQueryString(
//              replica.server,
//              replica.getUnsigned("port"),
              replica.getString("bucket"), //chooseBucket(replica)
              replica.rfn,
              expiration,
              this->pool_);

  return rloc;
}

void S3PoolHandler::doneWriting(const Location& loc) throw (DmException)
{
}

std::string S3PoolHandler::chooseBucket(const Replica& replica)
{
  const SecurityContext* context = this->stack->getSecurityContext();
  std::string vo = context->groups[0].name+this->pool_.bucketsalt();
  
  return vo;
}
