/// @file    plugins/s3/S3.h
/// @brief   plugin to store data in a s3 backend.
/// @author  Martin Philipp Hellmich <mhellmic@cern.ch>
#ifndef S3_H
#define	S3_H

#include <dmlite/cpp/dmlite.h>
#include <dmlite/cpp/pooldriver.h>
#include <dmlite/cpp/utils/security.h>
#include <dmlite/cpp/utils/urls.h>
#include <dmlite/cpp/catalog.h>

#include "S3Connection.h"
#include "S3Factory.h"

#define POOL_SPACE 5*1024*1024*1024 // 5 GByte - the maximum object size that amazon allows

namespace dmlite {

/// PoolDriver
class S3PoolDriver: public PoolDriver {
public:
  S3PoolDriver(S3Factory *factory) throw (DmException);

  ~S3PoolDriver();

  std::string getImplId() const throw ();
  
  void setStackInstance(StackInstance*) throw (DmException);
  void setSecurityContext(const SecurityContext*) throw (DmException);

  PoolHandler* createPoolHandler(const std::string& poolName) throw (DmException);
  
  void toBeCreated(const Pool& pool) throw (DmException);
  void justCreated(const Pool& pool) throw (DmException);
  void update(const Pool& pool) throw (DmException);
  void toBeDeleted(const Pool& pool) throw (DmException);

private:
  friend class S3PoolHandler;

  StackInstance* stack; 

  S3Factory*     factory_;
};

/// PoolHandler
class S3PoolHandler: public PoolHandler {
public:
  S3PoolHandler(S3Factory *factory, const std::string& poolName,
               StackInstance* si, S3PoolDetails pool);
  ~S3PoolHandler();

  std::string getPoolType(void) throw (DmException);
  std::string getPoolName(void) throw (DmException);
  uint64_t getTotalSpace(void) throw (DmException);
  uint64_t getFreeSpace(void) throw (DmException);
  bool poolIsAvailable(bool) throw (DmException);
  
  /// Check, if the replica exists on S3.
  /// If the replica has status 'Pending', it checks if the upload has
  /// been completed successfully.
  /// If it is marked as deleted, it return false, if '-' (active), it
  /// return true.
  /// Any other status returns false.
  /// @param replica      The file replica.
  /// @return             Whether the file is available or not.
  bool replicaIsAvailable(const Replica&) throw (DmException);
  
  /// Return the location of the replica.
  /// This function returns the replica location to which the client can 
  /// connect directly. The location is a signed link to S3 valid for a given time.
  /// For compatibility with dmlite 0.3, this function performs a check, if the
  /// replica has been successfully uploaded. In later version, this functionality 
  /// will be called from outside and can be removed here.
  /// @param replica      The file replica.
  /// @return             The file location.
  Location whereToRead (const Replica&) throw (DmException);
  void removeReplica   (const Replica&) throw (DmException);
  
  Location whereToWrite (const std::string&) throw (DmException);
  void doneWriting(const Location& loc) throw (DmException);


private:
  S3Factory*     factory_;

  std::string    bucket_;
  std::string    host_;
  unsigned int   port_;
  S3PoolDetails  pool_;

  /// The S3 Connection.
  S3Connection*  conn_;
  std::string    poolName; 
  StackInstance* stack;  

  /// Chooses the bucket by extracting the VO from the rfn.
  /// @param replica      The replica.
  /// @return             The bucket name.
  std::string chooseBucket(const Replica& replica);
};

/// IO Factory

};
#endif	// S3_H
