/// @file    plugins/s3/S3Connection.cpp
/// @brief   s3 plugin.
/// @author  Martin Philipp Hellmich <mhellmic@cern.ch>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "S3Connection.h"

#define CHECK_REPLICA_DELAY 10

using namespace dmlite;

boost::mutex S3Connection::replicaSetMutex;
boost::mutex S3Connection::stackMySqlMutex;
boost::mutex S3Connection::setStackMutex;

std::set<ReplicaMeta> S3Connection::pendingReplicas;

bool ReplicaMeta::operator <  (const ReplicaMeta& r) const
{
  return this->ingest_time < r.ingest_time;
}

S3Connection::S3Connection(/*const std::string& host, unsigned int port,
               const std::string& s3AccessKeyID,
               const std::string& s3SecretAccessKey,*/
               const time_t replicaTimeout):
 /*host(host), port(port),*/ replicaTimeout(replicaTimeout),
 stack(0x00)
{
  this->driver = S3Driver(); 
  this->startAvailabilityCheck();
}

S3Connection::~S3Connection()
{
  this->stopAvailabilityCheck();
  delete this->stack;
}

void S3Connection::setStackInstance(StackInstance *si) throw (DmException)
{
  boost::lock_guard<boost::mutex> lock(setStackMutex);
  if (this->stack == 0x00) {
    PluginManager *pm = si->getPluginManager();
    this->stack = new StackInstance(pm);
    if (this->stack == 0x00)
      throw DmException(DMLITE_SYSERR(ENOMEM),
                        "Could not create StackInstance for S3Connection");

    this->stack->setSecurityContext(*(si->getSecurityContext()));
  }
}

Location S3Connection::getQueryString(/*std::string host, unsigned int port,*/ std::string bucket, std::string key,
                          time_t expirationDate, S3PoolDetails pool)
{
  return this->driver.getQueryString("GET",/* host, pool.backend_port(),*/ bucket, key, expirationDate, pool);
}

Location S3Connection::putQueryString(/*std::string host, unsigned int port,*/ std::string bucket, std::string key,
                          time_t expirationDate, S3PoolDetails pool)
{
  return this->driver.getQueryString("PUT", /*pool.host(), pool.backend_port(),*/ bucket, key, expirationDate, pool);
}

S3RequestResponse S3Connection::removeReplica(Replica replica, S3PoolDetails pool)
{
  S3RequestResponse response = 
    this->driver.removeObject(/*replica.server, replica.getUnsigned("port")*//*pool.host(), pool.backend_port(),*/ replica.getString("bucket"), replica.rfn, pool);
 
 /*
  9.7 DELETE
  A successful response SHOULD be 200 (OK) if the response includes an entity
  describing the status, 202 (Accepted) if the action has not yet been
  enacted, or 204 (No Content) if the action has been enacted but the 
  response does not include an entity.
  http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html
 */
  if (response.http_code() == 200 ||
      response.http_code() == 202 ||
      response.http_code() == 204) {
    {
      boost::lock_guard<boost::mutex> lock(stackMySqlMutex);
      this->stack->getINode()->deleteReplica(replica);
    }
  } else {
    throw DmException(DMLITE_SYSERR(DMLITE_UNKNOWN_ERROR),
                      "%d %s %s",
                      response.http_code(),
                      response.http_reason().c_str(),
                      response.s3_reason().c_str());
  }

  return response;
}
  
void S3Connection::startAvailabilityCheck() throw (DmException)
{
  int errno;
  errno = pthread_create(&(this->checkThread), NULL, checkAvailabilityLoopEntry, this);
  if (errno) {
    throw DmException(DMLITE_SYSERR(errno), "S3 Check Thread creation failed");
  }
}

void S3Connection::stopAvailabilityCheck()
{
  int errno;
  void *joinRetval;
  errno = pthread_cancel(this->checkThread);
  if (errno) {
    throw DmException(DMLITE_SYSERR(errno), "S3 Check Thread submitting the cancellation failed");
  }
  // make sure the thread is canceled correctly
  errno = pthread_join(this->checkThread, &joinRetval);
  if (joinRetval != PTHREAD_CANCELED)
    throw DmException(DMLITE_SYSERR(errno), "S3 Check Thread cancel failed");

}

bool S3Connection::checkReplicaAvailability(Replica replica, S3PoolDetails pool)
{
  bool isAvailable = false;
  S3ObjectMetadata meta;
  S3RequestResponse response;
  response = 
     this->driver.headObject(/*replica.server, replica.getUnsigned("port") pool.backend_port(),*/ replica.getString("bucket"), replica.rfn, pool);

 /*
  9.6 PUT
  If an existing resource is modified, either the 200 (OK) or 204 (No Content)
  response codes > SHOULD be sent to indicate successful completion of 
  the request.
  http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html

  We need the returned content, so we only look for 200 (OK)
 */
  if (response.http_code() == 200) {
    meta = response.s3object_meta();
    // if the response was successful (file complete), change the db entry
    if (meta.has_content_length()) {
      replica.status = Replica::kAvailable;
      {
        boost::lock_guard<boost::mutex> lock(stackMySqlMutex);
        this->stack->getINode()->updateReplica(replica);
        this->stack->getINode()->setSize(replica.fileid, meta.content_length());
        this->stack->getINode()->setChecksum(replica.fileid, meta.checksum_type(), meta.checksum());
      }
      isAvailable = true;
    }
  }
  return isAvailable;
}

void *S3Connection::checkAvailabilityLoop()
{
  std::set<ReplicaMeta>::iterator itReplicas;
  time_t now = time(NULL);

  ReplicaMeta replicaMeta;
  //  Replica replica; // dont need it anymore
  bool noReplicasPending;

  while (true) {
    if (this->stack) {
      // check if something is available
      {
        boost::lock_guard<boost::mutex> lock(replicaSetMutex);
        noReplicasPending = pendingReplicas.empty();
      }

      // get replica
      if (!noReplicasPending) {
        {
          boost::lock_guard<boost::mutex> lock(replicaSetMutex);
          itReplicas = pendingReplicas.begin();
          replicaMeta = *itReplicas;
          pendingReplicas.erase(itReplicas);
        }

        // check for timeout
        now = time(NULL);
        if ((now - replicaMeta.ingest_time) > this->replicaTimeout) {
          // delete replica, if necessary
          {
            boost::lock_guard<boost::mutex> lock(stackMySqlMutex);
            this->stack->getINode()->deleteReplica(replicaMeta.replica);
          }
        } else {
          // check, if it is available, otherwise
          if (!checkReplicaAvailability(replicaMeta.replica, replicaMeta.pool)) {
            {
              boost::lock_guard<boost::mutex> lock(replicaSetMutex);
              pendingReplicas.insert(replicaMeta);
            }
          }
        }
      }
      // the only cancellation point in this thread
      sleep(CHECK_REPLICA_DELAY);
    }
  }
}

void *S3Connection::checkAvailabilityLoopEntry(void *This)
{
  int errno;
  int oldstate, oldtype;
  errno = pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oldstate);
  if (errno) {
    throw DmException(DMLITE_SYSERR(errno), "S3 Check Thread set cancel state failed");
  }
  // the thread can be only canceled at its cancellation points
  errno = pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &oldtype);
  if (errno) {
    throw DmException(DMLITE_SYSERR(errno), "S3 Check Thread set cancel type failed");
  }

  static_cast<S3Connection *>(This)->checkAvailabilityLoop();
  return NULL;
}

void S3Connection::addPendingReplica(const Replica& replica, S3PoolDetails pool)
{
  ReplicaMeta replicaMeta;
  replicaMeta.replica = replica;
  replicaMeta.ingest_time = time(NULL);
  replicaMeta.pool = pool;
  {
    boost::lock_guard<boost::mutex> lock(replicaSetMutex);
    pendingReplicas.insert(replicaMeta);
  }
}

