/// @file    plugins/s3/S3.h
/// @brief   s3 plugin.
/// @author  Martin Philipp Hellmich <mhellmic@cern.ch>

#include <vector>

#include "S3.h"
#include "S3Factory.h"

using namespace dmlite;

S3ConnectionFactory::S3ConnectionFactory(const std::string& host, unsigned int port):
/*host_(host), port_(port), s3AccessKeyID_(std::string()), s3SecretAccessKey_(std::string()),*/
s3ReplicaTimeout_(60*60*5)
{
  // Nothing
}

S3ConnectionFactory::~S3ConnectionFactory()
{
  // Nothing
}

S3Connection* S3ConnectionFactory::create()
{
  S3Connection *conn = new S3Connection(/*host_, port_, s3AccessKeyID_,
                                        s3SecretAccessKey_,*/ s3ReplicaTimeout_);

  return conn;
}

void S3ConnectionFactory::destroy(S3Connection* conn)
{
  delete conn;
}

bool S3ConnectionFactory::isValid(S3Connection* conn)
{
  // TODO
  return true;
}
/*
void S3ConnectionFactory::setS3SecretAccessKey(const std::string& key)
{
  this->s3SecretAccessKey_ = key;
}
*/
void S3ConnectionFactory::setS3ReplicaTimeout(std::string duration)
{
  time_t timeout = 0;

  unsigned int delta;
  std::stringstream converterStream;
  static const boost::regex durationRegex("(([[:digit:]]+)m)?(([[:digit:]]+)w)?(([[:digit:]]+)d)?(([[:digit:]]+)h)?([[:digit:]]+)?");
  boost::cmatch match;
  
  if (boost::regex_match(duration.c_str(), match, durationRegex)) {
    if (match[2].matched) {
      converterStream << std::string(match[2].first, match[2].second);
      converterStream >> delta;
      timeout += delta * (60*60*24*30); // add a month (30 days)
    }
    if (match[4].matched) {
      converterStream << std::string(match[4].first, match[4].second);
      converterStream >> delta;
      timeout += delta * (60*60*24*7); // add a week
    }
    if (match[6].matched) {
      converterStream << std::string(match[6].first, match[6].second);
      converterStream >> delta;
      timeout += delta * (60*60*24); // add a day
    }
    if (match[8].matched) {
      converterStream << std::string(match[8].first, match[8].second);
      converterStream >> delta;
      timeout += delta * (60*60); // add an hour
    }
    if (match[9].matched) {
      converterStream << std::string(match[9].first, match[9].second);
      converterStream >> delta;
      timeout += delta; // add seconds
    }
  } else {
    throw DmException(DMLITE_CFGERR(EINVAL),
        "Timeout string '%s' does not match the format", 
        duration.c_str());
  }
  
  this->s3ReplicaTimeout_ = timeout;
}

S3Factory::S3Factory() throw (DmException):
  connectionFactory_(std::string("s3.amazonaws.com"), 80),
  connectionPool_(&connectionFactory_, 2)
{
  pthread_key_create(&this->thread_s3_conn_, NULL);
}

S3Factory::~S3Factory() throw (DmException)
{
  pthread_key_delete(this->thread_s3_conn_);
}

void S3Factory::configure(const std::string& key, const std::string& value) throw (DmException)
{
  /*
  if (key == "S3Host") {
    this->connectionFactory_.host_ = value;
  } else if (key == "S3Port") {
    this->connectionFactory_.port_ = atoi(value.c_str());
  } else if (key == "S3BucketSalt") {
    this->bucketSalt_ = value;
  } else */ if (key == "S3Timeout") {
    try {
      this->connectionFactory_.setS3ReplicaTimeout(value);
    } catch (DmException e) {
      throw DmException(DMLITE_CFGERR(EINVAL),
          "Option %s recognised, but value %s has wrong format",
          key.c_str(), value.c_str());
    }
  } /* else if (key == "S3AccessKeyID") {
    this->connectionFactory_.s3AccessKeyID_ = value;
  } else if (key == "S3SecretAccessKey") {
    this->connectionFactory_.setS3SecretAccessKey(value);
  } */ else {
    throw DmException(DMLITE_CFGERR(DMLITE_UNKNOWN_KEY),
                      "Option %s not recognised", key.c_str());
  }
}

std::string S3Factory::implementedPool() throw ()
{
  return "s3";
}

PoolDriver* S3Factory::createPoolDriver() throw (DmException)
{
  return new S3PoolDriver(this);
}

S3Connection* S3Factory::getConnection(void) throw (DmException)
{
  S3Connection* conn;
  
  // Try from the thread local storage
  conn = static_cast<S3Connection*>(pthread_getspecific(this->thread_s3_conn_));
  
  // If NULL, instantiate and set in the thread local storage
  if (conn == NULL) {
    conn = this->connectionPool_.acquire();
    pthread_setspecific(this->thread_s3_conn_, conn);
  }
  // Increase reference count otherwise
  else {
    this->connectionPool_.acquire(conn);
  }
  
  // Return
  return conn;
}

void S3Factory::releaseConnection(S3Connection* conn) throw (DmException)
{
  if (this->connectionPool_.release(conn) == 0)
    pthread_setspecific(this->thread_s3_conn_, NULL);
}


static void registerPluginS3(PluginManager* pm) throw (DmException)
{
  pm->registerPoolDriverFactory(static_cast<PoolDriverFactory*>(new S3Factory()));
}

/// This is what the PluginManager looks for
PluginIdCard plugin_s3_pooldriver = {
  API_VERSION,
  registerPluginS3
};

